##############################################################################
#
#    Redner Odoo module
#    Copyright © 2016 XCG Consulting <https://xcg-consulting.fr>
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation, either version 3 of the
#    License, or (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
##############################################################################

import base64
import logging

from odoo import _, api, fields, models
from odoo.exceptions import ValidationError

from ..redner import REDNER_API_PATH, Redner

logger = logging.getLogger(__name__)

_redner = None

LANGUAGE_MJML_MUSTACHE = "text/mjml|mustache"
DEFAULT_LANGUAGE = LANGUAGE_MJML_MUSTACHE
COMPUTED_FIELDS = [
    "body",
    "description",
    "template_data",
    "slug",
    "is_mjml",
    "language",
    "locale_id",
]
EDITABLE_FIELDS = [
    "redner_id",
    "name",
]


class RednerTemplate(models.Model):
    _name = "redner.template"
    _description = "Redner Template"

    cache = {}

    source = fields.Selection(
        string="Source",
        selection=[
            ("redner", "Redner"),
            ("odoo", "Odoo"),
        ],
        required=True,
        default="odoo",
        help=("Source of the template, created in Odoo or imported from Redner."),
    )

    name = fields.Char(
        string="Name",
        help="This is a name of template mjml redner",
    )

    description = fields.Char(
        string="Description",
        help="Description of the template",
        readonly=False,
        compute="_compute_template",
    )

    body = fields.Text(
        string="Template body",
        translate=True,
        help="Code for the mjml redner template must be added here",
        readonly=False,
        compute="_compute_template",
    )

    slug = fields.Char(
        string="Slug",
        readonly=False,
        compute="_compute_template",
    )

    active = fields.Boolean(
        string="Active",
        default=True,
        help=(
            "If unchecked, it will allow you to hide the "
            "template without removing it."
        ),
    )

    is_mjml = fields.Boolean(
        string="Is MJML",
        default=True,
        help="set to false if your template doesn't contain MJML",
        readonly=False,
        compute="_compute_template",
    )

    detected_keywords = fields.Text(
        string="Variables", readonly=True, compute="_compute_keywords"
    )

    language = fields.Selection(
        string="Language",
        selection=[
            ("text/html|mustache", "HTML + mustache"),
            (LANGUAGE_MJML_MUSTACHE, "MJML + mustache"),
            (
                "application/vnd.oasis.opendocument.text|od+mustache",
                "OpenDocument + mustache",
            ),
        ],
        default="text/html|mustache",
        help="templating language",
        readonly=False,
        compute="_compute_template",
    )

    redner_id = fields.Char(string="Redner ID", readonly=True)

    locale_id = fields.Many2one(
        comodel_name="res.lang",
        string="Locale",
        help="Optional translation language (ISO code).",
        readonly=False,
        default=lambda self: self.env.ref("base.lang_fr").id,
        compute="_compute_template",
    )

    template_data = fields.Binary(
        string="Libreoffice Template",
        readonly=False,
        compute="_compute_template",
    )

    def import_from_redner(self):
        print("importing from redner ..")
        tl_wizard = self.env["template.list.wizard"]
        templates = self.list_external_templates()
        tl_wizard.populate(templates)

        return {
            "type": "ir.actions.act_window",
            "res_model": "template.list.wizard",
            "view_mode": "tree",
            "view_id": self.env.ref("redner.view_template_list_wizard_tree").id,
            "target": "new",
            #'context': self.env.context,
        }

    def _compute_template(self):
        print(f"compute_template. cache : {self.cache}")
        for record in self:
            if not record.id or not record.redner_id:
                continue
            if not any([getattr(record, f) for f in COMPUTED_FIELDS]):
                # if all the fields are undefined, get from redner
                try:
                    template = self.redner.templates.account_template_read(
                        record.redner_id
                    )
                    cached_template = self.to_cache(record.id, template)
                    for f in COMPUTED_FIELDS + EDITABLE_FIELDS:
                        if f in cached_template:
                            setattr(record, f, cached_template[f])
                except Exception as e:
                    logger.error("Failed to update redner template :%s", e)
                    return

            else:
                # if at least one field is defined, we get from the cache
                # if the field is undefined
                cached_template = self.cache[record.id]
                for f in COMPUTED_FIELDS:
                    attr = getattr(record, f)
                    if not attr:
                        setattr(record, f, cached_template[f])
        print(f"computed_template. cache : {self.cache}")

    def list_external_templates(self):
        try:
            template_list = self.redner.templates.account_template_list()
        except Exception as e:
            logger.error("Failed to list redner templates :%s", e)
            return []
        # Get the IDs of the templates that already exist in Odoo
        existing_template_ids = (
            self.env["redner.template"].search([]).mapped("redner_id")
        )
        print(f"existing templates {existing_template_ids}")
        new_templates = []
        for template in template_list:
            # Filter out templates that already exist in Odoo
            if template["name"] not in existing_template_ids:
                new_templates.append(self.to_odoo_template(template))
        return new_templates

    def to_cache(self, record_id, template):
        cached_template = self.to_odoo_template(template)
        self.cache[record_id] = cached_template
        return cached_template

    def to_odoo_template(self, template):
        # define a odoo_like template to put into the cache
        language = "{}|{}".format(template.get("produces"), template.get("language"))

        odoo_template = {
            "name": template.get("name"),
            "description": template.get("description", ""),
            "redner_id": template.get("name"),
            "locale_id": self.env["res.lang"].search(
                [("code", "=", template.get("locale", "fr_FR"))], limit=1
            ),
            "language": language,
            "slug": template.get("slug"),
            "is_mjml": language == LANGUAGE_MJML_MUSTACHE,
            "body": "",
            "template_data": False,
        }
        match template.get("body-format"):
            case "base64":
                body = base64.b64decode(template.get("body", ""))
            case _:
                body = template.get("body", "")
        if template.get("language") == "od+mustache":
            odoo_template["template_data"] = body
        else:
            odoo_template["body"] = body
        return odoo_template

    @property
    def redner(self):
        """Try to avoid Redner instance to be over created"""
        global _redner
        if _redner is None:
            # Bypass security rules when reading these configuration params. By
            # default, only some administrators have access to that model.
            config_model = self.env["ir.config_parameter"].sudo()

            _redner = Redner(
                config_model.get_param("redner.api_key"),
                config_model.get_param("redner.server_url"),
                config_model.get_param("redner.account"),
                int(config_model.get_param("redner.timeout", default="20")),
            )

        return _redner

    @api.model_create_multi
    def create(self, vals_list):
        """Overwrite create to create redner template"""

        for vals in vals_list:
            if self.env.context.get("importing"):
                return super().create(vals_list)

            # Prepare template params according to the selected language.
            # Use template data field if the selected language is "od";
            # otherwise the body field is used.
            produces, language = vals.get("language", DEFAULT_LANGUAGE).split("|")
            body, body_format = (
                (vals.get("template_data", ""), "base64")
                if language == "od+mustache"
                else (vals.get("body"), "text")
            )

            locale = self.env["res.lang"].browse(vals.get("locale_id")).code

            # We depend on the API for consistency here
            # So raised error should not result with a created template
            template = self.redner.templates.account_template_add(
                language=language,
                body=body,
                name=vals.get("name"),
                description=vals.get("description"),
                produces=produces,
                body_format=body_format,
                version=fields.Datetime.to_string(fields.Datetime.now()),
                locale=locale,
            )
            vals["redner_id"] = template["name"]

        return super().create(vals_list)

    def write(self, vals):
        """Overwrite write to update redner template"""
        # We depend on the API for consistency here
        # So raised error should not result with an updated template

        # pdb.set_trace()
        print(f"writing... {vals}\n")
        if "name" in vals:
            self.ensure_one()

            redner_id = self.redner_id
            vals["redner_id"] = vals["name"]

        ret = super().write(vals)
        should_update_redner = False
        for f in EDITABLE_FIELDS + COMPUTED_FIELDS:
            # if we made a change in the record, update redner
            if f in vals:
                attr = getattr(self, f)
                if vals[f] != attr:
                    should_update_redner = True

        for record in self:
            if should_update_redner and not self.env.context.get("importing"):
                try:
                    produces, language = record.language.split("|")
                    body, body_format = (
                        (record.template_data.decode(), "base64")
                        if language == "od+mustache"
                        else (record.body, "text")
                    )

                    if "name" not in vals:
                        redner_id = record.redner_id

                    record.redner.templates.account_template_update(
                        template_id=redner_id,
                        language=language,
                        body=body,
                        name=vals.get("name", ""),
                        description=record.description,
                        produces=produces,
                        body_format=body_format,
                        version=fields.Datetime.to_string(record.write_date),
                        locale=record.locale_id.code,
                    )
                except Exception as e:
                    logger.error("Failed to update redner template :%s", e)
                    raise ValidationError(
                        _("Failed to update render template, %s") % e
                    ) from e
        return ret

    def unlink(self):
        """Overwrite unlink to delete redner template"""

        # We do NOT depend on the API for consistency here
        # So raised error should not result block template deletion
        try:
            self.redner.templates.account_template_delete(self.redner_id)
        except Exception:
            pass

        return super().unlink()

    def copy(self, default=None):
        self.ensure_one()
        default = dict(default or {}, name=_("%s (copy)") % self.name)
        return super().copy(default)

    @api.depends("body", "template_data")
    def _compute_keywords(self):
        for record in self:
            record.detected_keywords = "\n".join(record.template_varlist_fetch())

    @api.model
    def get_keywords(self):
        """Return template redner keywords"""

        varlist = self.template_varlist_fetch()

        for name in varlist:
            while "." in name:
                name = name[: name.rfind(".")]
                if name not in varlist:
                    varlist.append(name)

        varlist.sort()

        return varlist

    @api.model
    def template_varlist_fetch(self):
        """Retrieve the list of variables present in the template."""
        try:
            if not self.redner_id:
                return []

            return self.redner.templates.account_template_varlist(self.redner_id)

        except Exception as e:
            logger.warning("Failed to fetch account template varlist: %s", e)
            return []

    def redner_url(self):
        if self.redner_id is None:
            return ""

        return (
            self.redner.server_url.removesuffix(REDNER_API_PATH)
            + "template/"
            + self.redner.account
            + "/"
            + self.redner_id
        )

    def view_in_redner(self):
        return {
            "type": "ir.actions.act_url",
            "url": self.redner_url(),
            "target": "new",
        }
